/*
 * $Id: GBTMips.c,v 1.1.1.1 2003/04/01 13:52:52 jkglee Exp $
 *
 * Flash and EPROM on 7751RGN.
 *
 * (C) 2001 GIGABYTE, Inc.
 *
 * GPL'd
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/config.h>

#define WINDOW_ADDR CONFIG_MTD_GBTMIPSMAP_START
#define WINDOW_SIZE CONFIG_MTD_GBTMIPSMAP_LEN
#define BUSWIDTH CONFIG_MTD_GBTMIPSMAP_BUSWIDTH

extern int parse_redboot_partitions(struct mtd_info *master, struct mtd_partition **pparts);

__u16 mips_GBTMips_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

void mips_GBTMips_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void mips_GBTMips_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}


static struct mtd_info *flash_mtd;
static struct mtd_info *eprom_mtd;

static struct mtd_partition *parsed_parts;

struct map_info mips_GBTMips_flash_map = {
	name: "GIGABYTE MIPS FLASH",
	size: 0x100000,
	buswidth: 2,
	read16: mips_GBTMips_read16,
	copy_from: mips_GBTMips_copy_from,
	write16: mips_GBTMips_write16,
};


#ifdef CONFIG_MTD_GBT_RESERVE
static struct mtd_partition mips_GBTMips_partitions[] = {
	/* Reserved for boot code, read-only */
	{
		name: "flash_boot",
		offset: 0x00000000,
		size: CONFIG_MTD_GBT_RESERVE,
		// by jkglee, 2002.11.25 to make this area writeable ...
		//mask_flags: MTD_WRITEABLE,
	},
	// by jkglee 2002.09.26
	{
		name: "data_setting",
		//offset: MTDPART_OFS_NXTBLK,
		//offset: CONFIG_MTD_SUPERH_RESERVE,
		//offset: MTDPART_OFS_APPEND,//copy from ts.hwang
		offset: CONFIG_MTD_GBT_RESERVE+CONFIG_MTD_MAC_SIZE,
		size: CONFIG_MTD_DATA_SETTING,
	},	
	/* All else is writable (e.g. JFFS) */
	{
		name: "kernel",
		offset: MTDPART_OFS_NXTBLK,
		//offset: CONFIG_MTD_SUPERH_RESERVE,
		size: CONFIG_MTD_GBT_KERNEL_SIZE,
	},

#if 0
	{
		name: "file_system",
		offset: MTDPART_OFS_NXTBLK,
		size: MTDPART_SIZ_FULL,	
	}
#else
	{
		name: "file_system",
		offset: MTDPART_OFS_NXTBLK,
		size: CONFIG_MTD_ROOTFS,
	},
#if 1	
	{
		name: "default_mac",
		offset: CONFIG_MTD_GBT_RESERVE,
		size: CONFIG_MTD_MAC_SIZE,	
	}
#endif
#endif	
	
	

};
#endif /* CONFIG_MTD_GBT_RESERVE */

static int __init init_GBTMips_maps(void)
{
	int nr_parts = 0;

	/* First probe at offset 0 */
	mips_GBTMips_flash_map.map_priv_1 = (unsigned long)ioremap(WINDOW_ADDR, WINDOW_SIZE);;
	if (!mips_GBTMips_flash_map.map_priv_1) {
                printk("Failed to ioremap\n");
                return -EIO;
        }
	
	printk(KERN_NOTICE "Probing for flash chips at 0x00000000 by using cfi_probe:\n");
  	//flash_mtd = do_map_probe("jedec_probe", &mips_GBTMips_flash_map);
	flash_mtd = do_map_probe("cfi_probe", &mips_GBTMips_flash_map);
	if (!flash_mtd) {
		/* Not there. Try swapping */
		printk(KERN_NOTICE "Probing for flash chips at 0x00000000 by using jedec_probe:\n");
		//mips_GBTMips_flash_map.map_priv_1 = P2SEGADDR(0x01000000);
		//mips_GBTMips_eprom_map.map_priv_1 = P1SEGADDR(0);
		flash_mtd = do_map_probe("jedec_probe", &mips_GBTMips_flash_map);
		if (!flash_mtd) {
			/* Eep. */
			printk(KERN_NOTICE "Flash chips not detected at either possible location.\n");
			return -ENXIO;
		}
	}
	printk(KERN_NOTICE "physmap flash device: %x at %x\n", WINDOW_SIZE, WINDOW_ADDR);
	flash_mtd->module = THIS_MODULE;
#ifdef CONFIG_MTD_REDBOOT_PARTS
	nr_parts = parse_redboot_partitions(flash_mtd, &parsed_parts);
	if (nr_parts > 0)
		printk(KERN_NOTICE "Found RedBoot partition table.\n");
	else if (nr_parts < 0)
		printk(KERN_NOTICE "Error looking for RedBoot partitions.\n");
#endif /* CONFIG_MTD_REDBOOT_PARTS */
#if CONFIG_MTD_GBT_RESERVE
	if (nr_parts == 0) {
		printk(KERN_NOTICE "Using configured partition at 0x%08x.\n",
		       CONFIG_MTD_GBT_RESERVE);
		parsed_parts = mips_GBTMips_partitions;
		nr_parts = sizeof(mips_GBTMips_partitions)/sizeof(*parsed_parts);
	}
#endif /* CONFIG_MTD_GBT_RESERVE */

	if (nr_parts > 0)
		add_mtd_partitions(flash_mtd, parsed_parts, nr_parts);
	else
		add_mtd_device(flash_mtd);

	return 0;
}

static void __exit cleanup_GBTMips_maps(void)
{
#if 0
	if (eprom_mtd) {
		del_mtd_device(eprom_mtd);
		map_destroy(eprom_mtd);
	}
#endif
	if (parsed_parts)
		del_mtd_partitions(flash_mtd);
	else
		del_mtd_device(flash_mtd);
	map_destroy(flash_mtd);
}

module_init(init_GBTMips_maps);
module_exit(cleanup_GBTMips_maps);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("GIGABYTE");
MODULE_DESCRIPTION("MTD map driver for GIGABYTE MIPS platform");

