#include <linux/init.h>
#include <linux/sched.h>
#include <asm/irq.h>
#include <linux/mtd/mtd.h>
#include <linux/interrupt.h>
#include <linux/reboot.h>
#include <linux/sched.h>
#include <linux/tqueue.h>
#include <linux/gigabyte_gpio.h> /* Frank, 01Apr2003 */

int ResetConf_flag = 0;

static struct tq_struct ResetConf_task;

void ResetConf_mtd_erase_callback (struct erase_info *instr)
{
	 wake_up((wait_queue_head_t *)instr->priv);
};

void ResetConf_sched(unsigned long unused)
{
	int ret = 0;
	struct mtd_info *mtd_if;

	struct erase_info *erase=kmalloc(sizeof(struct erase_info),GFP_KERNEL);
	u_int32_t start_addr=0;
	mtd_if = __get_mtd_device(NULL,1); // get mtd 1
        if (!erase)
	{
		printk("malloc memory fail\n");
		ResetConf_flag = 0;
		return;
	}else
	{
		// erase each block
		for (start_addr = 0;start_addr < mtd_if->size;start_addr += mtd_if->erasesize)
		{
			wait_queue_head_t waitq;
                       	DECLARE_WAITQUEUE(wait, current);
                                      
                       	init_waitqueue_head(&waitq);
                     
                       	memset (erase,0,sizeof(struct erase_info));

			erase->addr = start_addr;
			erase->len = mtd_if->erasesize;

			erase->mtd = mtd_if;
                       	erase->callback = ResetConf_mtd_erase_callback;
                       	erase->priv = (unsigned long)&waitq;
				
			printk("Erase MTD 1 Address:%X Size:%d KBytes\n",erase->addr,erase->len/1024);

			/*
		  	FIXME: Allow INTERRUPTIBLE. Which means
		  	not having the wait_queue head on the stack.

                  	If the wq_head is on the stack, and we
		  	leave because we got interrupted, then the
		  	wq_head is no longer there when the
		  	callback routine tries to wake us up.
			*/
			ret = mtd_if->erase(mtd_if, erase);
			if (!ret) {
		      		set_current_state(TASK_UNINTERRUPTIBLE);
		      		add_wait_queue(&waitq, &wait);
		      		if (erase->state != MTD_ERASE_DONE &&
		          	erase->state != MTD_ERASE_FAILED)
				{
		               		schedule();
				}
		      		remove_wait_queue(&waitq, &wait);
		      		set_current_state(TASK_RUNNING);
                      		ret = (erase->state == MTD_ERASE_FAILED)?-EIO:0;
				if (ret == -EIO)
					printk("Erase %X (size %d KBytes)fail\n",erase->addr,erase->len/1024);
                       	}	
		}
                kfree(erase);
	}
	// reboot 
	machine_restart(NULL);
}

void ResetConf_interrupt(int irq, void *dev_id, struct pt_regs *egs)
{
	unsigned long flags;
       
	save_flags(flags);
	cli();
	if (ResetConf_flag == 0)
	{
		ResetConf_flag = 1;
		ResetConf_task.routine = ResetConf_sched;
		schedule_task(&ResetConf_task);
	}
	restore_flags(flags);
}

void __init ResetConf_init(void)
{
	int ret = 0;
	
/* Frank, 01Apr2003 : Added for using NMI as init default data interrupt for AP102AB */

	/*this is enable GPIO 10 as a input pin*/
  *(volatile unsigned long *)(0xbd010000)=*(volatile unsigned long *)(0xbd010000) | (0x00000002);
  *(volatile unsigned long *)(0xbd010040)=*(volatile unsigned long *)(0xbd010040) & (~(1<<10));
  *(volatile unsigned long *)(0xbd010048)=*(volatile unsigned long *)(0xbd010048) | (0x00300000);
   ret=request_irq(0x01,ResetConf_interrupt,SA_INTERRUPT,"Reset Configure Setting ISR",NULL);
	if (ret)
	{
		printk("ResetConf_init: request_irq fail %d\n",ret);
	}
}

__initcall(ResetConf_init);

