/* MSNP extension for IP connection tracking. */
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <linux/ctype.h>
#include <net/checksum.h>
#include <net/tcp.h>

#include <linux/netfilter_ipv4/lockhelp.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_msnp.h>

DECLARE_LOCK(ip_msnp_lock);
struct module *ip_conntrack_msnp = THIS_MODULE;

#define MAX_PORTS 8
static int ports[MAX_PORTS];
static int ports_c;
#ifdef MODULE_PARM
MODULE_PARM(ports, "1-" __MODULE_STRING(MAX_PORTS) "i");
#endif

static int loose = 0;
MODULE_PARM(loose, "i");

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif


#define ADDR_LINE_LEN	(17)	/* xxx.xxx.xxx.xxx\r\n */
#define PORT_LINE_LEN	(7)		/* xxxxx\r\n */

#define INVITE_FT_PORT_MIN	(6891)
#define INVITE_FT_PORT_MAX	(6900)

static int try_number(const char *data, size_t dlen, u_int32_t array[],
		      int array_size, char sep, char term)
{
	u_int32_t i, len;

	memset(array, 0, sizeof(array[0])*array_size);

	/* Keep data pointing at next char. */
	for (i = 0, len = 0; len < dlen && i < array_size; len++, data++) {
		if (*data >= '0' && *data <= '9') {
			array[i] = array[i]*10 + *data - '0';
		}
		else if (*data == sep)
			i++;
		else {
			/* Unexpected character; true if it's the
			   terminator and we're finished. */
			if (*data == term && i == array_size - 1)
				return len;

			DEBUGP("Char %u (got %u nums) `%u' unexpected\n",
			       len, i, *data);
			return 0;
		}
	}
	DEBUGP("Failed to fill %u numbers separated by %c\n", array_size, sep);

	return 0;
}



/* Return 1 for match, 0 for accept, -1 for partial. */
static int find_pattern(const char *data, size_t dlen,
			unsigned int *numoff,
			unsigned int *numlen,
			u_int32_t array[5])
{
	size_t i;

	const char* strAddr = "IP-Address: ";
	const char* strPort = "Port: ";
	const char* strAuthCookie = "AuthCookie: ";
	const int	strAddr_len = strlen(strAddr);
	const int	strPort_len = strlen(strPort);
	const int	strAuth_len = strlen(strAuthCookie);
	char*	p_addr;
	char*	p_port;
	char*	p_auth;
	
	if(dlen <= strAddr_len + strPort_len + strAuth_len){
		return 0;	/* No match */
	}
	
	/* Search String "IP-Address: xxx.xxx.xxx.xxx\r\nPort: xxxxx\r\nAuth-Cookie: "*/
	p_addr = strstr(data, strAddr);
	p_port = strstr(data, strPort);
	p_auth = strstr(data, strAuthCookie);

	if(p_addr == NULL || p_port == NULL || p_auth == NULL){
		return 0;	/* No match */
	}
	printk(" IP-Address,Port,Auth-Cookie Field exist\n");
	if(!(p_addr < p_port && p_addr + strAddr_len+ ADDR_LINE_LEN >= p_port
		&& p_port < p_auth && p_port + strPort_len+ PORT_LINE_LEN >= p_auth)){
		return 0;	/* No match */
	}
	
	*numoff = i = p_addr + strAddr_len - data;	/* index of address region*/
	
	if(!try_number(data + i, ADDR_LINE_LEN, array, 4, '.', '\r'))
	{
		return 0;
	}
	

	if(!try_number(p_port + strPort_len , PORT_LINE_LEN, array+4, 1, '.', '\r')) /* '.' is dummy */
	{

		return 0;
	}
	
	if(array[4] < INVITE_FT_PORT_MIN || INVITE_FT_PORT_MAX < array[4])
	{
		return 0;
	}
	
	*numlen = p_auth - p_addr - strAddr_len;
	
	DEBUGP("Match succeeded!\n");
	return 1;
}

/* Return 1 for match, 0 for accept, -1 for partial. */
static int find_pattern_msn5(const char *data, size_t dlen,
			unsigned int *numoff,
			unsigned int *numlen,
			u_int32_t array[11])
{
	size_t i;

	const char* strAddr = "IP-Address: ";
	const char* strAddrInt = "IP-Address-Internal: ";
	const char* strPortXInt = "PortX-Internal: ";
	const char* strPort = "Port: ";
	const char* strPortX = "PortX: ";
	const char* strAuthCookie = "AuthCookie: ";
	const int	strAddr_len = strlen(strAddr);
	const int	strAddrInt_len	= strlen(strAddrInt);
	const int	strPortXInt_len = strlen(strPortXInt);
	const int	strPort_len = strlen(strPort);
	const int	strPortX_len = strlen(strPortX);
	const int	strAuth_len = strlen(strAuthCookie);
	char*	p_addr;
	char*	p_addrInt;
	char* 	p_portXInt;
	char*	p_port;
	char*	p_portX;
	char*	p_auth;
	
	if(dlen <= strAddr_len + strPort_len + strPortX_len + strAuth_len){
		DEBUGP("MSN5: Length is short..\n");
		return 0;	/* No match */
	}
	
	/* Search String "IP-Address: xxx.xxx.xxx.xxx\r\nPort: xxxxx\r\nAuth-Cookie: "*/
	p_addr = strstr(data, strAddr);
	p_addrInt = strstr(data, strAddrInt);
	p_portXInt = strstr(data, strPortXInt);
	p_port = strstr(data, strPort);
	p_portX = strstr(data, strPortX);
	p_auth = strstr(data, strAuthCookie);

	/*
	if (p_portXInt) {
		DEBUGP("MSN5:FT use UPnP..\n");
		return 0;
	}
	*/
	
	if(p_addr == NULL || p_port == NULL || p_portX == NULL || p_auth == NULL){
		DEBUGP("MSN5:No Match\n");
		return 0;	/* No match */
	}
	
	printk("MSN5:Match! FT ALG USE\n");
	
	*numoff = i = p_addr + strAddr_len - data;	/* index of address region*/
	
	if(!try_number(data + i, ADDR_LINE_LEN, array, 4, '.', '\r'))
	{
		return 0;
	}
	
	if (p_addrInt) {
		DEBUGP("MSN5:IP-Address-Internal exist\n");
		if(!try_number(p_addrInt + strAddrInt_len, ADDR_LINE_LEN, array+5, 4, '.', '\r'))
		{
			DEBUGP("IP-Address-Internal error\n");
			return 0;
		}
	} else {
		/* If "IP-Address-Internal" is not found. array[5] ~ array[8] equals to 0 */
		array[5] = 0;
		array[6] = 0;
		array[7] = 0;
		array[8] = 0;
	}
	if (p_portXInt) {
		DEBUGP("MSN5:PortX-Internal exist\n");
		if(!try_number(p_portXInt + strPortXInt_len , PORT_LINE_LEN, array+10, 1, '.', '\r')) /* '.' is dummy */
		{
			DEBUGP("PortX-Internal:error\n");
			return 0;
		}
	} else {
		array[10] = 0;
	}
	
	if(!try_number(p_port + strPort_len , PORT_LINE_LEN, array+4, 1, '.', '\r')) /* '.' is dummy */
	{

		return 0;
	}
	
	if(!try_number(p_portX + strPortX_len , PORT_LINE_LEN, array+9, 1, '.', '\r')) /* '.' is dummy */
	{
		DEBUGP("PortX:error\n");
		return 0;
	}
	
	if(array[4] < INVITE_FT_PORT_MIN)
	{
		DEBUGP("MSN5:PORT < 6891\n");
		return 0;
	}
	
	*numlen = p_auth - p_addr - strAddr_len;
	
	DEBUGP("MSN5:Match succeeded!\n");
	return 1;
}


/* FIXME: This should be in userspace.  Later. */
static int help(const struct iphdr *iph, size_t len,
		struct ip_conntrack *ct,
		enum ip_conntrack_info ctinfo)
{
	/* tcplen not negative guaranteed by ip_conntrack_tcp.c */
	struct tcphdr *tcph = (void *)iph + iph->ihl * 4;
	const char *data = (const char *)tcph + tcph->doff * 4;
	unsigned int tcplen = len - iph->ihl * 4;
	unsigned int datalen = tcplen - tcph->doff * 4;
	u_int32_t old_seq_aft_nl;
	int old_seq_aft_nl_set;
	u_int32_t array[10] = { 0 };
	int dir = CTINFO2DIR(ctinfo);
	unsigned int matchlen, matchoff;
	struct ip_conntrack_tuple t, mask;
	struct ip_ct_msnp *info = &ct->help.ct_msnp_info;
	int found = 0;
	int version;

	/* Until there's been traffic both ways, don't look in packets. */
	if (ctinfo != IP_CT_ESTABLISHED
	    && ctinfo != IP_CT_ESTABLISHED+IP_CT_IS_REPLY) {
		DEBUGP("msnp: Conntrackinfo = %u\n", ctinfo);
		return NF_ACCEPT;
	}

	/* Not whole TCP header? */
	if (tcplen < sizeof(struct tcphdr) || tcplen < tcph->doff*4) {
		DEBUGP("msnp: tcplen = %u\n", (unsigned)tcplen);
		return NF_ACCEPT;
	}

	/* Checksum invalid?  Ignore. */
	/* FIXME: Source route IP option packets --RR */
	if (tcp_v4_check(tcph, tcplen, iph->saddr, iph->daddr,
			 csum_partial((char *)tcph, tcplen, 0))) {
		DEBUGP("msnp_help: bad csum: %p %u %u.%u.%u.%u %u.%u.%u.%u\n",
		       tcph, tcplen, NIPQUAD(iph->saddr),
		       NIPQUAD(iph->daddr));
		return NF_ACCEPT;
	}

	LOCK_BH(&ip_msnp_lock);
	old_seq_aft_nl_set = info->seq_aft_nl_set[dir];
	old_seq_aft_nl = info->seq_aft_nl[dir];

	DEBUGP("conntrack_msnp: datalen %u\n", datalen);
	if (datalen > 0) {
		DEBUGP("conntrack_msnp: datalen %u ends in \\n\n", datalen);
		if (!old_seq_aft_nl_set
		    || after(ntohl(tcph->seq) + datalen, old_seq_aft_nl)) {
			DEBUGP("conntrack_msnp: updating nl to %u\n",
			       ntohl(tcph->seq) + datalen);
			info->seq_aft_nl[dir] = ntohl(tcph->seq) + datalen;
			info->seq_aft_nl_set[dir] = 1;
		}
	}
	UNLOCK_BH(&ip_msnp_lock);

	if(!old_seq_aft_nl_set ||
			(ntohl(tcph->seq) != old_seq_aft_nl)) {
		DEBUGP("ip_conntrack_msnp_help: wrong seq pos %s(%u)\n",
		       old_seq_aft_nl_set ? "":"(UNSET) ", old_seq_aft_nl);
		return NF_ACCEPT;
	}

	if(dir !=IP_CT_DIR_ORIGINAL){
		return NF_ACCEPT;
	}
	
	found = find_pattern_msn5(data, datalen, &matchoff, &matchlen, array);
	
	if (found == 1) {
		version = 5;
	} else {
		found = find_pattern(data, datalen, &matchoff,&matchlen, array);
		
		if (found == 1) {
			version = 4;
		} else { /* No Match */
			return NF_ACCEPT;
		}
	}

	DEBUGP("conntrack_msnp: match `%.*s' (%u bytes at %u)\n",
	       (int)matchlen, data + matchoff,
	       matchlen, ntohl(tcph->seq) + matchoff);

	/* Update the msnp info */
	LOCK_BH(&ip_msnp_lock);
	
	if (version == 5) {
		info->is_msnp = 1863;
		info->seq = ntohl(tcph->seq) + matchoff;
		info->len = matchlen;
		info->port = array[4];
		info->version = 5;
		info->msnp_data.ip = ((array[3] << 24) | (array[2] << 16) | (array[1] << 8) | array[0]);
		info->msnp_data.ip_internal = ((array[8] << 24) | (array[7] << 16) | (array[6] << 8) | array[5]);
		info->msnp_data.portX_internal = array[10];
		info->msnp_data.port = array[4];
		DEBUGP("ip_conntrack_msnp::array[9]: %u\n", array[9]);
		info->msnp_data.portX = array[9];
		DEBUGP("ip_conntrack_msnp::info->msnp_data.portX: %u\n", info->msnp_data.portX);
	} else if (htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3])
	    == ct->tuplehash[dir].tuple.src.ip) {
		info->is_msnp = 1863;
		info->seq = ntohl(tcph->seq) + matchoff;
		info->len = matchlen;
		info->port = array[4];
	    info->version = 4;
	} else {
		DEBUGP("conntrack_msnp: NOT RECORDING: %u,%u,%u,%u != %u.%u.%u.%u\n",
		       array[0], array[1], array[2], array[3],
		       NIPQUAD(ct->tuplehash[dir].tuple.src.ip));

		/* Thanks to Cristiano Lincoln Mattos
		   <lincoln@cesar.org.br> for reporting this potential
		   problem (DMZ machines opening holes to internal
		   networks, or the packet filter itself). */
		if (!loose) goto out;
	}

	t = ((struct ip_conntrack_tuple)
		{ { ct->tuplehash[!dir].tuple.src.ip,
		    { 0 } },
		  { htonl((array[0] << 24) | (array[1] << 16)
			  | (array[2] << 8) | array[3]),
		    { htons(array[4]) },
		    IPPROTO_TCP }});
	mask = ((struct ip_conntrack_tuple)
		{ { 0x0, { 0 } },
		  { 0xFFFFFFFF, { 0xFFFF }, 0xFFFF }});
	/* Ignore failure; should only happen with NAT */
	ip_conntrack_expect_related(ct, &t, &mask, NULL);
 out:
	UNLOCK_BH(&ip_msnp_lock);

	return NF_ACCEPT;
}

static struct ip_conntrack_helper msnp[MAX_PORTS];

/* Not __exit: called from init() */
static void fini(void)
{
	int i;
	for (i = 0; (i < MAX_PORTS) && ports[i]; i++) {
		DEBUGP("ip_ct_msnp: unregistering helper for port %d\n",
				ports[i]);
		ip_conntrack_helper_unregister(&msnp[i]);
	}
}

static int __init init(void)
{
	int i, ret;

	if (ports[0] == 0)
		ports[0] = 1863;

	for (i = 0; (i < MAX_PORTS) && ports[i]; i++) {
		memset(&msnp[i], 0, sizeof(struct ip_conntrack_helper));
		msnp[i].tuple.src.u.tcp.port = htons(ports[i]);
		msnp[i].tuple.dst.protonum = IPPROTO_TCP;
		msnp[i].mask.src.u.tcp.port = 0xFFFF;
		msnp[i].mask.dst.protonum = 0xFFFF;
		msnp[i].help = help;
		DEBUGP("ip_ct_msnp: registering helper for port %d\n", 
				ports[i]);
		ret = ip_conntrack_helper_register(&msnp[i]);

		if (ret) {
			fini();
			return ret;
		}
		ports_c++;
	}
	return 0;
}


EXPORT_SYMBOL(ip_msnp_lock);
EXPORT_SYMBOL(ip_conntrack_msnp);

module_init(init);
module_exit(fini);
