/* AFS3CB extension for UDP NAT alteration. */

// Support for MSN Messenger 5.0 2002/11/07

#include <linux/module.h>
#include <linux/netfilter_ipv4.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <net/udp.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#define MAX_PORTS 8
static int ports[MAX_PORTS];
static int ports_c = 0;

#ifdef MODULE_PARM
MODULE_PARM(ports, "1-" __MODULE_STRING(MAX_PORTS) "i");
#endif

DECLARE_LOCK(ip_afs3cb_lock);

#define		MSGR_VER_4	0x01		// 03/2/13
#define		MSGR_VER_5	0x02		// 03/2/13
static 	int	Handle_ver = MSGR_VER_4;	// 03/2/13
MODULE_PARM     (Handle_ver, "i");		// 03/2/13
#if 1
 #define DEBUGP_ALG(x)	printk(x)
#endif

#define AFS3CB_PORT_OFFSET	(2)
#define AFS3CB_IP_OFFSET	(4)
int 
ip_nat_afs3cb_mangle_packet(struct sk_buff **skb,
			 u_int32_t pip,
			 u_int16_t pport)
{
	struct iphdr* iph = (*skb)->nh.iph;
	struct udphdr* udph;
	unsigned char* data;
	u_int32_t udplen, newlen, newudplen;

	newudplen = udplen = (*skb)->len - iph->ihl*4;
	newlen = iph->ihl*4 + newudplen;

	if (skb_cloned(*skb)) {
		struct sk_buff *nskb = skb_copy(*skb, GFP_ATOMIC);
		if (!nskb) {
			if (net_ratelimit())
				printk("Out of memory cloning UDP packet\n");
			return 0;
		}
		if ((*skb)->sk)
			skb_set_owner_w(nskb, (*skb)->sk);
		kfree_skb(*skb);
		*skb = nskb;
	}

	/* skb may be copied !! */
	iph = (*skb)->nh.iph;
	udph = (void *)iph + iph->ihl*4;
	data = (void *)udph + sizeof(struct udphdr);

	memcpy(data + AFS3CB_PORT_OFFSET, (char*) &pport, sizeof(pport));
	memcpy(data + AFS3CB_IP_OFFSET, (char*) &pip, sizeof(pip));
	
	// This function changes only UDP payload.
	// UDP header checksum must be re-calculated.
//	(*skb)->csum = csum_partial((char *)udph + sizeof(struct udphdr),
//				    udplen - sizeof(struct udphdr));

	if(udph->check)
	{
	  udph->check = 0;
	}
	ip_send_check(iph);

	return 1;
}



static unsigned int help(struct ip_conntrack *ct,
			 struct ip_nat_info *info,
			 enum ip_conntrack_info ctinfo,
			 unsigned int hooknum,
			 struct sk_buff **pskb)
{
	struct iphdr *iph = (*pskb)->nh.iph;
	struct udphdr *udph = (void *)iph + iph->ihl*4;
	const char *data = (const char *) udph + sizeof(struct udphdr);
	u_int32_t* pip = (u_int32_t*) (data + AFS3CB_IP_OFFSET);
	u_int16_t* pport = (u_int16_t*) (data + AFS3CB_PORT_OFFSET);

	unsigned int datalen;
	int dir;


	/* Only mangle things once: original direction in POST_ROUTING
	   and reply direction on PRE_ROUTING. */
	dir = CTINFO2DIR(ctinfo);
	if (!(hooknum == NF_IP_PRE_ROUTING && dir == IP_CT_DIR_REPLY)) {
		DEBUGP("nat_afs3cb: Not touching dir %s at hook %s\n",
		       dir == IP_CT_DIR_ORIGINAL ? "ORIG" : "REPLY",
		       hooknum == NF_IP_POST_ROUTING ? "POSTROUTING"
		       : hooknum == NF_IP_PRE_ROUTING ? "PREROUTING"
		       : hooknum == NF_IP_LOCAL_OUT ? "OUTPUT" : "???");
		return NF_ACCEPT;
	}

	datalen = (*pskb)->len - iph->ihl * 4 - sizeof(struct udphdr);
	LOCK_BH(&ip_afs3cb_lock);
	if (datalen >= 8)	/* 01(or 02) 02 pp pp aa aa aa aa */
	{
		if(*(data+1) == 0x02)
		{
			switch(*data)
			{
				case 0x01:
					// Messenger 4.7
				   	if( Handle_ver & MSGR_VER_4 ) // 03/2/13
				   	{
					if(*pport == ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u.udp.port 
						&& *pip == ct ->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip)
					{
						ip_nat_afs3cb_mangle_packet(pskb, 
							ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip,
							ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port);
						DEBUGP_ALG("pc2phone ALG:4.7\n\r");
					}
				   	}
					break;
				case 0x02:
					// Messenger 5.0
				   	if( Handle_ver & MSGR_VER_5 ) // 03/2/13
				   	{
					if(((*pport)^htons(0x4131)) == ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u.udp.port 
						&& ((*pip)^htonl(0x41314131)) == ct ->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip)
					{
						ip_nat_afs3cb_mangle_packet(pskb, 
							ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip ^ htonl(0x41314131),
							ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port ^ htons(0x4131));
						DEBUGP_ALG("pc2phone ALG:5.0\n\r");
					}
					}
					break;
				default:
					break;
			}
		}
	}

	UNLOCK_BH(&ip_afs3cb_lock);

	return NF_ACCEPT;
}

static struct ip_nat_helper afs3cb[MAX_PORTS];
static char afs3cb_names[MAX_PORTS][9];


/* Not __exit: called from init() */

static void fini(void)
{
	int i;

	for (i = 0; (i < MAX_PORTS) && ports[i]; i++) {
		DEBUGP("ip_nat_afs3cb: unregistering port %d\n", ports[i]);
		ip_nat_helper_unregister(&afs3cb[i]);
	}
}

static int __init init(void)
{
	int i, ret;
	char *tmpname;

	printk("pc2phone ALG: parameter=%x \n\r",Handle_ver);

	if (ports[0] == 0)
		ports[0] = 7001;

	for (i = 0; (i < MAX_PORTS) && ports[i]; i++) {

		memset(&afs3cb[i], 0, sizeof(struct ip_nat_helper));

		afs3cb[i].tuple.dst.protonum = IPPROTO_UDP;
		afs3cb[i].tuple.src.u.udp.port = htons(ports[i]);
		afs3cb[i].mask.dst.protonum = 0xFFFF;
		afs3cb[i].mask.src.u.udp.port = 0xFFFF;
		afs3cb[i].help = help;

		tmpname = &afs3cb_names[i][0];
		sprintf(tmpname, "afs3cb%2.2d", i);
		afs3cb[i].name = tmpname;

		DEBUGP("ip_nat_afs3cb: Trying to register for port %d\n",
			ports[i]);
		ret = ip_nat_helper_register(&afs3cb[i]);
		if (ret) {
			printk("ip_nat_afs3cb: error registering helper for port %d\n", ports[i]);
			fini();
			return ret;
		}
		ports_c++;
	}
	return ret;
}

module_init(init);
module_exit(fini);

