/* MSNP extension for TCP NAT alteration. */
#include <linux/module.h>
#include <linux/netfilter_ipv4.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <net/tcp.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ip_conntrack_msnp.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#define DEF_PORT	6891		/* default port Messenger FT use */
#define PORT2PORTX_NUM		4288 /* MSN Messenger ver5 use this num to translate Port to PortX */

#define MAX_PORTS 8

static u_int16_t port2portx(u_int16_t port, const struct ip_ct_msnp *ct_msnp_info);

static int ports[MAX_PORTS];
static int ports_c = 0;

#ifdef MODULE_PARM
MODULE_PARM(ports, "1-" __MODULE_STRING(MAX_PORTS) "i");
#endif

DECLARE_LOCK_EXTERN(ip_msnp_lock);

/* FIXME: Time out? --RR */

static int
msnp_nat_expected(struct sk_buff **pskb,
		 unsigned int hooknum,
		 struct ip_conntrack *ct,
		 struct ip_nat_info *info,
		 struct ip_conntrack *master,
		 struct ip_nat_info *masterinfo,
		 unsigned int *verdict)
{
	struct ip_nat_multi_range mr;
	u_int32_t newdstip, newsrcip, newip;
	struct ip_ct_msnp *msnpinfo;

	IP_NF_ASSERT(info);
	IP_NF_ASSERT(master);
	IP_NF_ASSERT(masterinfo);

	IP_NF_ASSERT(!(info->initialized & (1<<HOOK2MANIP(hooknum))));

	DEBUGP("nat_expected: We have a connection!\n");
	/* Master must be an msnp connection */
	msnpinfo = &master->help.ct_msnp_info;

	LOCK_BH(&ip_msnp_lock);
	//if (msnpinfo->is_msnp != 1863 || HOOKMAINIP(hooknum) == IP_NAT_MANIP_SRC) {
	if (msnpinfo->is_msnp != 1863 || HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC) {
		UNLOCK_BH(&ip_msnp_lock);
		DEBUGP("nat_expected: master not msnp\n");
		return 0;
	}

	newdstip = master->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
	newsrcip = master->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip;
	DEBUGP("nat_expected: MSNP FT %u.%u.%u.%u->%u.%u.%u.%u\n",
	       NIPQUAD(newsrcip), NIPQUAD(newdstip));
	UNLOCK_BH(&ip_msnp_lock);

	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC)
		newip = newsrcip;
	else
		newip = newdstip;

	DEBUGP("nat_expected: IP to %u.%u.%u.%u\n", NIPQUAD(newip));


	mr.rangesize = 1;
	/* We don't want to manip the per-protocol, just the IPs... */
	mr.range[0].flags = IP_NAT_RANGE_MAP_IPS;
	mr.range[0].min_ip = mr.range[0].max_ip = newip;

	/* ... unless we're doing a MANIP_DST, in which case, make
	   sure we map to the correct port */
	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_DST) {
		mr.range[0].flags |= IP_NAT_RANGE_PROTO_SPECIFIED;
		mr.range[0].min = mr.range[0].max
			= ((union ip_conntrack_manip_proto)
				{ htons(msnpinfo->port) });
	}
	*verdict = ip_nat_setup_info(ct, &mr, hooknum);

	return 1;
}

static int
mangle_packet(struct sk_buff **pskb,
		     u_int32_t newip,
		     u_int16_t port,
		     unsigned int matchoff,
		     unsigned int matchlen,
		     struct ip_conntrack *ct,
		     enum ip_conntrack_info ctinfo)
{
	char buffer[sizeof("xxx.xxx.xxx.xxx\r\nPort: xxxxx\r\n")];

	MUST_BE_LOCKED(&ip_msnp_lock);

	sprintf(buffer, "%u.%u.%u.%u\r\nPort: %u\r\n",
		NIPQUAD(newip), port);

	DEBUGP("calling ip_nat_mangle_tcp_packet\n");

	return ip_nat_mangle_tcp_packet(pskb, ct, ctinfo, matchoff, 
					matchlen, buffer, strlen(buffer));
}

static int mangle_packet_msn5(struct sk_buff **pskb,
		     u_int32_t ip,
			 u_int32_t ip_internal,
			 u_int16_t portXInt,
		     u_int16_t port,
			 u_int16_t portX,
		     unsigned int matchoff,
		     unsigned int matchlen,
		     struct ip_conntrack *ct,
		     enum ip_conntrack_info ctinfo)
{
	char buffer[sizeof("xxx.xxx.xxx.xxx\r\nIP-Address-Internal: xxx.xxx.xxx.xxx\r\nPortX-Internal: xxxxx\r\nPort: xxxxx\r\nPortX: xxxxx\r\n")];
	char tmp_buf[50];
	MUST_BE_LOCKED(&ip_msnp_lock);
	
	sprintf(buffer, "%u.%u.%u.%u\r\n", NIPQUAD(ip));
	if (ip_internal) {
		DEBUGP("IP-Address-Internal add;\n");
		sprintf(tmp_buf, "IP-Address-Internal: %u.%u.%u.%u\r\n", NIPQUAD(ip_internal));
		strcat(buffer, tmp_buf);
	}
	if (portXInt) {
		DEBUGP("portX-Internal add;\n");
		sprintf(tmp_buf, "PortX-Internal: %u\r\n", portXInt);
		strcat(buffer, tmp_buf);
	}
	sprintf(tmp_buf, "Port: %u\r\n", port);
	strcat(buffer, tmp_buf);
	sprintf(tmp_buf, "PortX: %u\r\n", portX);
	strcat(buffer, tmp_buf);

	
	DEBUGP("calling ip_nat_mangle_tcp_packet\n");

	return ip_nat_mangle_tcp_packet(pskb, ct, ctinfo, matchoff, 
						matchlen, buffer, strlen(buffer));
}

static int msnp_data_fixup(const struct ip_ct_msnp *ct_msnp_info,
			  struct ip_conntrack *ct,
			  unsigned int datalen,
			  struct sk_buff **pskb,
			  enum ip_conntrack_info ctinfo)
{
	u_int32_t newip;
	struct iphdr *iph = (*pskb)->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl*4;
	u_int16_t port;
	struct ip_conntrack_tuple tuple;
	/* Don't care about source port */
	const struct ip_conntrack_tuple mask
		= { { 0x0, { 0 } },
		    { 0xFFFFFFFF, { 0xFFFF }, 0xFFFF } };

	memset(&tuple, 0, sizeof(tuple));
	MUST_BE_LOCKED(&ip_msnp_lock);
	DEBUGP("MSNP_NAT: seq %u + %u in %u + %u\n",
	       ct_msnp_info->seq, ct_msnp_info->len,
	       ntohl(tcph->seq), datalen);

	/* Change address inside packet to match way we're mapping
	   this connection. */
	newip = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
	/* Expect something from server->client */
	tuple.src.ip = ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip;
	tuple.dst.ip = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;

	tuple.dst.protonum = IPPROTO_TCP;

	/* Try to get same port: if not, try to change it. */
	for (port = ct_msnp_info->port; port != 0; port++) {
		tuple.dst.u.tcp.port = htons(port);

		if (ip_conntrack_expect_related(ct, &tuple, &mask, NULL) == 0)
			break;
	}
	if (port == 0)
		return 0;

	if (ct_msnp_info->version == 5) {
		if (!mangle_packet_msn5(pskb, newip, ct_msnp_info->msnp_data.ip_internal, ct_msnp_info->msnp_data.portX_internal, port, port2portx(port, ct_msnp_info),
						  ct_msnp_info->seq - ntohl(tcph->seq),
						  ct_msnp_info->len, ct, ctinfo))
			return 0;
	} else {
		if (!mangle_packet(pskb, newip, port,
						  ct_msnp_info->seq - ntohl(tcph->seq),
						  ct_msnp_info->len, ct, ctinfo))
			return 0;
	}

	return 1;
}

static unsigned int help(struct ip_conntrack *ct,
			 struct ip_nat_info *info,
			 enum ip_conntrack_info ctinfo,
			 unsigned int hooknum,
			 struct sk_buff **pskb)
{
	struct iphdr *iph = (*pskb)->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl*4;
	unsigned int datalen;
	int dir;
	int score;
	struct ip_ct_msnp *ct_msnp_info
		= &ct->help.ct_msnp_info;

	/* Delete SACK_OK on initial TCP SYNs. */
	if (tcph->syn && !tcph->ack)
		ip_nat_delete_sack(*pskb, tcph);

	/* Only mangle things once: original direction in POST_ROUTING
	   and reply direction on PRE_ROUTING. */
	dir = CTINFO2DIR(ctinfo);
	if (!((hooknum == NF_IP_POST_ROUTING && dir == IP_CT_DIR_ORIGINAL)
	      || (hooknum == NF_IP_PRE_ROUTING && dir == IP_CT_DIR_REPLY))) {
		DEBUGP("nat_msnp: Not touching dir %s at hook %s\n",
		       dir == IP_CT_DIR_ORIGINAL ? "ORIG" : "REPLY",
		       hooknum == NF_IP_POST_ROUTING ? "POSTROUTING"
		       : hooknum == NF_IP_PRE_ROUTING ? "PREROUTING"
		       : hooknum == NF_IP_LOCAL_OUT ? "OUTPUT" : "???");
		return NF_ACCEPT;
	}

	datalen = (*pskb)->len - iph->ihl * 4 - tcph->doff * 4;
	score = 0;
	LOCK_BH(&ip_msnp_lock);
	if (ct_msnp_info->len) {
		/* If it's in the right range... */
		score += between(ct_msnp_info->seq, ntohl(tcph->seq),
				 ntohl(tcph->seq) + datalen);
		score += between(ct_msnp_info->seq + ct_msnp_info->len,
				 ntohl(tcph->seq),
				 ntohl(tcph->seq) + datalen);
		if (score == 1) {
			/* Half a match?  This means a partial retransmisison.
			   It's a cracker being funky. */
			if (net_ratelimit()) {
				printk("MSNP_NAT: partial packet %u/%u in %u/%u\n",
				       ct_msnp_info->seq, ct_msnp_info->len,
				       ntohl(tcph->seq),
				       ntohl(tcph->seq) + datalen);
			}
			UNLOCK_BH(&ip_msnp_lock);
			return NF_DROP;
		} else if (score == 2) {
			if (!msnp_data_fixup(ct_msnp_info, ct, datalen,
					    pskb, ctinfo)) {
				UNLOCK_BH(&ip_msnp_lock);
				return NF_DROP;
			}
			/* skb may have been reallocated */
			iph = (*pskb)->nh.iph;
			tcph = (void *)iph + iph->ihl*4;
		}
	}

	UNLOCK_BH(&ip_msnp_lock);

	ip_nat_seq_adjust(*pskb, ct, ctinfo);

	return NF_ACCEPT;
}

static struct ip_nat_helper msnp[MAX_PORTS];
static char msnp_names[MAX_PORTS][7];

static struct ip_nat_expect msnp_expect
= { { NULL, NULL }, msnp_nat_expected };

/* Not __exit: called from init() */
static void fini(void)
{
	int i;

	for (i = 0; (i < MAX_PORTS) && ports[i]; i++) {
		DEBUGP("ip_nat_msnp: unregistering port %d\n", ports[i]);
		ip_nat_helper_unregister(&msnp[i]);
	}

	ip_nat_expect_unregister(&msnp_expect);
}

static int __init init(void)
{
	int i, ret;
	char *tmpname;

	ret = ip_nat_expect_register(&msnp_expect);
	if (ret == 0) {
		if (ports[0] == 0)
			ports[0] = 1863;

		for (i = 0; (i < MAX_PORTS) && ports[i]; i++) {

			memset(&msnp[i], 0, sizeof(struct ip_nat_helper));

			msnp[i].tuple.dst.protonum = IPPROTO_TCP;
			msnp[i].tuple.src.u.tcp.port = htons(ports[i]);
			msnp[i].mask.dst.protonum = 0xFFFF;
			msnp[i].mask.src.u.tcp.port = 0xFFFF;
			msnp[i].help = help;

			tmpname = &msnp_names[i][0];
			sprintf(tmpname, "msnp%2.2d", i);
			msnp[i].name = tmpname;

			DEBUGP("ip_nat_msnp: Trying to register for port %d\n",
					ports[i]);
			ret = ip_nat_helper_register(&msnp[i]);

			if (ret) {
				printk("ip_nat_msnp: error registering helper for port %d\n", ports[i]);
				fini();
				return ret;
			}
			ports_c++;
		}

	} else {
		ip_nat_expect_unregister(&msnp_expect);
	}
	return ret;
}

static u_int16_t port2portx(u_int16_t port, const struct ip_ct_msnp *ct_msnp_info) {
	
	/* if def_port, not change portX */
	if (port == DEF_PORT) {
		DEBUGP("port2portx#1::return: %u", ct_msnp_info->msnp_data.portX);
		return ct_msnp_info->msnp_data.portX;
	}
	
	if ((port % 2) == 0) {
		DEBUGP("port2portx#2::return: %u\n", port + PORT2PORTX_NUM + 1);
		return (port + PORT2PORTX_NUM + 1);
	} else {
		DEBUGP("port2portx#3::return: %u\n", port + PORT2PORTX_NUM + 1);
		return (port + PORT2PORTX_NUM - 1);
	}
	
	return 0;
}

module_init(init);
module_exit(fini);
